from __future__ import unicode_literals

import errno
import functools
import mock
import os
import pathlib
import shutil
import subprocess
import tempfile
import unittest
import re

from six import BytesIO

import zc.buildout

import rubygems

RUBYGEMS_DEFAULT_VERSION = '2.0.0'


def touch(path):
    with path.open('w') as f:
        f.write('')


class fixture(object):
    bin_dir = os.path.join('rubygems', 'bin')
    rubygems_path = os.path.join(bin_dir, 'gem')

    def __init__(self, options=None, version=RUBYGEMS_DEFAULT_VERSION,
                 rubygems_installed=False):
        self.options = options or {}
        self.rubygems_installed = rubygems_installed
        self.version = self.options.get('return', {}).get('version', version)

        self.to_patch = (
            ('check_output', 'rubygems.subprocess.check_output'),
            ('clean_up', 'rubygems.Recipe._clean_up'),
            ('urlopen', 'rubygems.urllib.request.urlopen'),
            ('download', 'rubygems.Download'),
        )
  
        self.patch_return_value_dict = {
            'urlopen': BytesIO(bytes(
                          ('https://rubygems.org/rubygems/rubygems-%s.zip'
                            % RUBYGEMS_DEFAULT_VERSION).encode('utf-8')
                       )),
        }

        self.patch_side_effect_dict = {
            'check_output': self.check_output_mock,
        }

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(test):
            buildout, name, options, version = self.set_up()
            cwd = os.getcwd()
            os.chdir(str(self.tempdir))
            func(test, self.tempdir, self.patches, buildout, name, options, version)
            os.chdir(cwd)
            self.tear_down()
        return wrapper

    def check_output_mock(self, cmd, **kwargs):
        if cmd[:3] == ['ruby', 'setup.rb', 'all']:
            self.install_rubygems()

    def install_rubygems(self):
        touch(self.tempdir / self.rubygems_path)

    def patch(self, modules):
        self.patchers = {}
        self.patches = {}
        for name, module in modules:
            self.patchers[name] = mock.patch(module)
            self.patches[name] = self.patchers[name].start()

    def makedirs(self, dirs):
        self.tempdir = pathlib.Path(tempfile.mkdtemp())
        for directory in dirs:
            os.makedirs(str(self.tempdir / directory))

    def set_up(self):
        name = 'rubygems'
 
        self.patch(self.to_patch)
        for k, v in self.patch_return_value_dict.items():
            self.patches[k].return_value = v
        for k, v in self.patch_side_effect_dict.items():
            self.patches[k].side_effect = v

        self.makedirs((
            'bin',
            'ruby-' + self.version,
            'rubygems-' + self.version,
            self.bin_dir,
        ))
        if self.rubygems_installed:
            self.install_rubygems()

        buildout = {'buildout': dict({
            'parts-directory': str(self.tempdir),
            'bin-directory': str(self.tempdir / 'bin'),
        }, **self.options.get('buildout', {}))}

        options = self.options.get('recipe', {})

        return buildout, name, options, self.version

    def tear_down(self):
        for patcher in self.patchers.values():
            patcher.stop()
        shutil.rmtree(str(self.tempdir))


class RubyGemsTestCase(unittest.TestCase):
    def check_output_test(self, check_output_mock, expected_arg_list_list):
        self.assertEqual(check_output_mock.call_count,
                          len(expected_arg_list_list))

        for mock_call, expected_arg_list in zip(check_output_mock.mock_calls,
                                                expected_arg_list_list):
            self.assertEqual(mock_call[1][0], expected_arg_list)

    def install_with_default_rubygems_test(self, path, patches,
                                            expected_install_arg_list_list):
        # One urlopen call to get latest version
        self.assertEqual(patches['urlopen'].call_count, 1)

        args = patches['urlopen'].mock_calls[0][1]
        self.assertEqual(args, ('https://rubygems.org/pages/download',))

        # One slapos.recipe.downloadunpacked call to download rubygems
        self.assertEqual(patches['download'].call_count, 1)

        args = patches['download'].mock_calls[0][1]
        self.assertEqual(args[2], {
            'url': 'https://rubygems.org/rubygems/rubygems-%s.zip'
                    % RUBYGEMS_DEFAULT_VERSION,
            'destination': str(path / ('rubygems-' + RUBYGEMS_DEFAULT_VERSION)),
        })

        expected_install_arg_list_list.insert(0, [
            'ruby', 'setup.rb', 'all', '--prefix=%s/rubygems' % path,
            '--no-rdoc', '--no-ri',
        ])
        self.check_output_test(
            patches['check_output'], expected_install_arg_list_list)


class RubyGemsDefaultTestCase(RubyGemsTestCase):
    @fixture({'recipe': {'gems': 'sass'}})
    def test_success(self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

        expected_install_arg_list_list = [
            [
                'ruby', str(path / fixture.rubygems_path), 'install',
                '--no-document', '--bindir=' + str(path / fixture.bin_dir),
                'sass', '--',
            ],
        ]
        self.install_with_default_rubygems_test(
            path, patches, expected_install_arg_list_list) 

    @fixture({'recipe': {}})
    def test_missing_gems(self, path, patches, buildout, name, options, version):
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            "Configuration error, 'gems' option is missing",
            rubygems.Recipe, buildout, name, options
        )

    @fixture({'recipe': {'gems': 'sass'}})
    def test_oserror(self, path, patches, buildout, name, options, version):
        patches['check_output'].side_effect = OSError
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'System error, command failed: .*',
            recipe.install
        )

    @fixture({'recipe': {'gems': 'sass'}})
    def test_signal_received(self, path, patches, buildout, name, options, version):
        exception = subprocess.CalledProcessError(-1, '')
        patches['check_output'].side_effect = exception
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'System error, command received signal 1: .*',
            recipe.install
        )

    @fixture({'recipe': {'gems': 'sass'}})
    def test_non_zero_exitcode(self, path, patches, buildout, name, options, version):
        exception = subprocess.CalledProcessError(1, '')
        patches['check_output'].side_effect = exception
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'System error, command failed with exit code 1: .*',
            recipe.install
        )

    @fixture({'recipe': {'gems': 'sass'}})
    def test_update(self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.update()

    @fixture({'recipe': {'gems': 'sass', 'environment': 'invalid'}})
    def test_invalid_environment(self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
          zc.buildout.UserError,
          'Configuration error, '
          'every environment line should contain a "=" sign',
          recipe.install
        )

    @fixture({'recipe': {
        'gems': 'sass',
        'url': 'https://rubygems.org/rubygems/rubygems-%s.zip'
                % RUBYGEMS_DEFAULT_VERSION,
    }})
    def test_version_from_url(self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

    @fixture({'recipe': {'gems': 'sass'}, 'version': RUBYGEMS_DEFAULT_VERSION})
    def test_version(self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

    @fixture({'recipe': {'gems': 'sass'}})
    def test_no_version(self, path, patches, buildout, name, options, version):
        patches['urlopen'].return_value = BytesIO(b'')
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(zc.buildout.UserError,
            'Can\'t find latest rubygems version.',
            recipe.install
        )

    @fixture({'recipe': {'gems': 'sass'}})
    @mock.patch('rubygems.os.mkdir')
    def test_mkdir_error(self, path, patches, buildout, name, options, version, mkdir):
        mkdir.side_effect = OSError(errno.EIO)
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'IO error while creating %s/rubygems directory.' % path,
            recipe.install
        )

    @fixture({'recipe': {'gems': 'sass'}})
    def test_executables(self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        touch(pathlib.Path(recipe.options['location']) / 'bin/sass')
        recipe.install()

    @fixture({'recipe': {'gems': 'sass==1.0'}, 'rubygems_installed': True})
    def test_pinned_versions(self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

        args = patches['check_output'].mock_calls[0][1]
        for arg in args[0]:
            matched = re.search('^--version.*$', arg)
            if matched:
                self.assertEqual(matched.group(), '--version=1.0')
                break

    @fixture({'recipe': {'gems': 'sass'}, 'version': '3.1.2'})
    def test_non_ascii_encoding(self, path, patches, buildout, name, options, version):
        rubygems_setup_output_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            'test_data', 
            'rubygems_3.1.2_setup_output.txt'
        )
        with open(rubygems_setup_output_path) as fd:
            rubygems_setup_output = fd.read()
        patches['check_output'].return_value = rubygems_setup_output

        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

    @fixture({'recipe': {
        'gems': 'sass',
        'url': 'https://rubygems.org/rubygems/rubygems-%s.zip'
                % RUBYGEMS_DEFAULT_VERSION,
        'version': RUBYGEMS_DEFAULT_VERSION,
    }})
    def test_version_and_url(self, path, patches, buildout, name, options, version):
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            "Configuration error, 'url' and 'version' options are mutually exclusive",
            rubygems.Recipe, buildout, name, options
        )

    @fixture({'recipe': {'gems': 'sass', 'version': '10.0.0'}})
    def test_2_digits_version(self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.update()


class deployment_fixture(fixture):
    def check_output_mock(self, cmd, **kwargs):
        super(deployment_fixture, self).check_output_mock(cmd)
        if len(cmd) > 3 and cmd[2] == 'dependency':
            filename = 'dependency_%s-%s.txt' % (cmd[5][2:-2], cmd[4])
            dependency_output_file = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), 'test_data', filename)
            with open(dependency_output_file) as fd:
                dependency_output = fd.read()

            return dependency_output

    def set_up(self):
        buildout, name, options, version = super(deployment_fixture, self).set_up()
        options['deployment'] = 'true'
        return buildout, name, options, version


class RubyGemsDeploymentTestCase(RubyGemsTestCase):
    def deployment_install_test(self, buildout, name, path, patches, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

        expected_install_arg_list_list = []
        gem_dict_list = list(map(recipe.get_gem_dict, recipe.gems))
        for gem_dict in gem_dict_list:
            gem_search_pattern = '^' + gem_dict['gemname'].replace('.',r'\.') + '$'
            if version[0] < '3':
                gem_search_pattern = '/' + gem_search_pattern + '/'

            gem_executable = str(path / fixture.rubygems_path)
            expected_install_arg_list_list.extend([
                [
                    'ruby', gem_executable, 'dependency', '-rv', gem_dict['version'],
                    gem_search_pattern,
                ], [
                    'ruby', gem_executable, 'install', '--no-document',
                    '--bindir=' + str(path / fixture.bin_dir),
                    '--ignore-dependencies', gem_dict['gemname'],
                    '--version=' + gem_dict['version'], '--',
                ]
            ])

        self.install_with_default_rubygems_test(
            path, patches, expected_install_arg_list_list)

    @deployment_fixture({
        'recipe': {'gems': 'hashie==0.3.1'},
        'rubygems_installed': True
    })
    def test_already_installed_rubygems(
            self, path, patches, buildout, name, options, version):
        self.deployment_install_test(
            buildout, name, path, patches, options, version)

    @deployment_fixture({'recipe': {'gems': 'sass'}})
    def test_deployment_not_pinned_version_error(
            self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version for gem sass is missing',
            recipe.install
        )

    @deployment_fixture({'recipe': {'gems': 'rash==0.1.0'}})
    def test_deployment_not_pinned_dependency_error(
            self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version for dependency hashie is missing',
            recipe.install
        )

    @deployment_fixture({'recipe': {'gems': 'rash==0.1.0 hashie==0.3.0'}})
    def test_deployment_more_or_equal_constraint_error(
            self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 0.3.0 for gem hashie '
            'does not satisfy dependency constraint >= 0.3.1 of gem rash',
            recipe.install
        )

    @deployment_fixture({'recipe': {'gems': 'rash==0.1.0 hashie==0.3.1'}})
    def test_deployment_more_or_equal_constraint_success(
            self, path, patches, buildout, name, options, version):
        self.deployment_install_test(
            buildout, name, path, patches, options, version)

    @deployment_fixture({'recipe': {'gems': 'rash==0.3.0 hashie==1.1.0'}})
    def test_deployment_similar_constraint_error(
            self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 1.1.0 for gem hashie '
            'does not satisfy dependency constraint < 1.1 of gem rash',
            recipe.install
        )

    @deployment_fixture({'recipe': {'gems': 'rash==0.3.0 hashie==1.0.0'}})
    def test_deployment_similar_constraint_success(
            self, path, patches, buildout, name, options, version):
        self.deployment_install_test(
            buildout, name, path, patches, options, version)

    @deployment_fixture({'recipe': {
        'gems': 'activesupport==3.2.0 atomic==0.0.1 cucumber-gherkin==9.1.0 '
                'cucumber-messages==9.0.2 i18n==0.6.0 json==2.3.0 '
                'middleware==0.1.0 multi_json==1.0.0 protobuf-cucumber==3.10.8 '
                'thor==0.9.2 thread_safe==0.1.0'
    }})
    def test_deployment_more_or_equal_constraint_error(
            self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 9.0.2 for gem cucumber-messages '
            'does not satisfy dependency constraint >= 9.0.3 '
            'of gem cucumber-gherkin',
            recipe.install
        )

    @deployment_fixture({'recipe': {
        'gems': 'activesupport==3.2.0 atomic==0.0.1 cucumber-gherkin==9.1.0 '
                'cucumber-messages==9.0.3 i18n==0.6.0 json==2.3.0 '
                'middleware==0.1.0 multi_json==1.0.0 protobuf-cucumber==3.10.8 '
                'thor==0.9.2 thread_safe==0.1.0'
    }})
    def test_deployment_more_or_equal_constraint_success(
            self, path, patches, buildout, name, options, version):
        self.deployment_install_test(
            buildout, name, path, patches, options, version)

    @deployment_fixture({'recipe': {'gems': 'hoe==3.4.0 rake==11.0.0'}})
    def test_deployment_less_constraint_error(
            self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 11.0.0 for gem rake '
            'does not satisfy dependency constraint < 11.0 of gem hoe',
            recipe.install
        )

    @deployment_fixture({'recipe': {'gems': 'hoe==3.4.0 rake==10.0.1'}})
    def test_deployment_less_constraint_success(
            self, path, patches, buildout, name, options, version):
        self.deployment_install_test(
            buildout, name, path, patches, options, version)

    @deployment_fixture({'recipe': {
        'gems': 'cucumber-core==3.0.0 backports==3.8.0 '
                'cucumber-tag_expressions==1.0.1 gherkin==8.0.0'
    }})
    def test_deployment_not_pinned_subdependency_error(
            self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version for dependency '
            'cucumber-messages is missing',
            recipe.install
        )

    @deployment_fixture({'recipe': {
        'gems': 'cucumber-core==3.0.0 backports==3.8.0 '
                'cucumber-tag_expressions==1.0.1 gherkin==8.0.0 '
                'cucumber-messages==6.0.1'
    }})
    def test_deployment_pinned_subdependency_success(
            self, path, patches, buildout, name, options, version):
        self.deployment_install_test(
            buildout, name, path, patches, options, version)

    @deployment_fixture({'recipe': {
        'gems': 'activesupport==3.2.0 atomic==0.0.1 cucumber-core==7.1.0 '
                'cucumber-messages==12.2.0 cucumber-gherkin==14.1.0 '
                'cucumber-tag-expressions==2.0.4 i18n==0.6.0 middleware==0.1.0 '
                'multi_json==1.0.0 protobuf-cucumber==3.10.8 thor==0.9.2 '
                'thread_safe==0.1.0'
    }})
    def test_deployment_subdependency_constraint_error(
            self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 12.2.0 for gem cucumber-messages '
            'does not satisfy dependency constraint >= 12.3 '
            'of gem cucumber-gherkin',
            recipe.install
        )

    @deployment_fixture({'recipe': {
        'gems': 'activesupport==3.2.0 atomic==0.0.1 cucumber-core==7.1.0 '
                'cucumber-messages==12.3.2 cucumber-gherkin==14.1.0 '
                'cucumber-tag-expressions==2.0.4 i18n==0.6.0 middleware==0.1.0 '
                'multi_json==1.0.0 protobuf-cucumber==3.10.8 thor==0.9.2 '
                'thread_safe==0.1.0'
    }})
    def test_deployment_subdependency_constraint_success(
            self, path, patches, buildout, name, options, version):
        self.deployment_install_test(
            buildout, name, path, patches, options, version)

    @deployment_fixture({'recipe': {
        'gems': 'activesupport==3.2.0 atomic==0.0.1 cucumber-core==7.1.0 '
                'cucumber-messages==12.3.2 cucumber-gherkin==15.0.0 '
                'cucumber-tag-expressions==2.0.4 i18n==0.6.0 middleware==0.1.0 '
                'multi_json==1.0.0 protobuf-cucumber==3.10.8 thor==0.9.2 '
                'thread_safe==0.1.0'
    }})
    def test_deployment_similar_second_constraint_error(
            self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 15.0.0 for gem cucumber-gherkin '
            'does not satisfy dependency constraint < 15 of gem cucumber-core',
            recipe.install
        )

    @deployment_fixture({'recipe': {
        'gems': 'activesupport==3.2.0 atomic==0.0.1 cucumber-core==7.1.0 '
                'cucumber-messages==12.3.2 cucumber-gherkin==14.1.0 '
                'cucumber-tag-expressions==2.0.4 i18n==0.6.0 middleware==0.1.0 '
                'multi_json==1.0.0 protobuf-cucumber==3.10.8 thor==0.9.2 '
                'thread_safe==0.1.0'
    }})
    def test_deployment_similar_second_constraint_success(
            self, path, patches, buildout, name, options, version):
        self.deployment_install_test(buildout, name, path, patches, options, version)
  
    @deployment_fixture({'recipe': {
        'gems': 'hashie==1.0.0 sass==1.0.0'
    }})
    def test_deployment_version_exclusion_constraint_error(
            self, path, patches, buildout, name, options, version):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 1.0.0 for gem hashie '
            'does not satisfy dependency constraint != 1.0.0 of gem sass',
            recipe.install
        )

    @deployment_fixture({'recipe': {
        'gems': 'hashie==0.3.1 sass==1.0.0'
    }})
    def test_deployment_version_exclusion_constraint_success(
            self, path, patches, buildout, name, options, version):
        self.deployment_install_test(buildout, name, path, patches, options, version)
